package org.LiveGraph.gui;

import java.awt.BorderLayout;
import java.awt.FontMetrics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollBar;
import javax.swing.WindowConstants;
import javax.swing.JScrollPane;
import java.awt.Dimension;
import java.io.File;
import java.util.Arrays;

import javax.swing.JTextArea;
import javax.swing.JLabel;
import javax.swing.JButton;

import javax.swing.JSlider;
import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;

import org.LiveGraph.LiveGraph;
import org.LiveGraph.dataCache.CacheObserver;
import org.LiveGraph.dataCache.DataCache;
import org.LiveGraph.dataCache.UpdateInvoker;
import org.LiveGraph.dataCache.UpdateInvokerObserver;
import org.LiveGraph.settings.DataFileSettings;
import org.LiveGraph.settings.ErrorWhileSettingHasChangedProcessingException;
import org.LiveGraph.settings.ObservableSettings;
import org.LiveGraph.settings.SettingsObserver;

import com.softnetConsult.utils.swing.SwingTools;

/**
 * The "Data File Settings" window of the application.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: DataFileSettingsWindow.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class DataFileSettingsWindow extends JFrame
									implements UpdateInvokerObserver, SettingsObserver, CacheObserver {

private JLabel intervalLabel = null;
private JTextArea fileInfoArea = null;
private JLabel fileNameLabel = null;
private JSlider updateIntervallSlider = null;
private JLabel nextUpdateLabel = null;
private JCheckBox dontCacheBox = null;
private JRadioButton showTailDataButton = null;
private JRadioButton showAllDataButton = null;
private JFileChooser openFileDialog = null;

private static final String [] updateIntervalLabels = {"every 1 second.",
													   "every 2 seconds.", "every 3 seconds.",
													   "every 5 seconds.", "every 10 seconds.",
													   "every 15 seconds.", "every 20 seconds.",													   
													   "every 30 seconds.", "every 45 seconds.",
													   "every 1 minute.", "every 90 seconds (1.5 minutes).",
													   "every 2 minutes.", "every 3 minutes.",
													   "every 5 minutes.", "every 10 minutes.",
													   "every 15 minutes.", "every 20 minutes.",
													   "every 30 minutes.", "every 45 minutes.",
													   "every 1 hour.", "only manual update."};
private static final long [] updateIntervalValues = {1000, 2000, 3000, 5000, 10000, 15000,
													 20000, 30000, 45000, 60000, 90000, 120000, 180000,
													 300000, 600000, 900000, 1200000, 1800000, 2700000,
													 3600000, -1};
static {
	if (updateIntervalLabels.length != updateIntervalValues.length)
		throw new Error("The arrays \"updateIntervalLabels\" and \"updateIntervalValues\" are not of the same size!");
}


/**
 * This is the default constructor.
 */
public DataFileSettingsWindow() {
	super();
	initialize();
}


/**
 * This method initializes the data file settings window.
 */
private void initialize() {
	
	// Window settings:
	
	final DataFileSettingsWindow DATAFILE_WIN = this;
	this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);	
	Dimension frameDim = new Dimension(470, 300);
	this.setPreferredSize(frameDim);
	this.setBounds(5, 5, frameDim.width, frameDim.height);	
	this.setTitle("Data file settings (LiveGraph)");
	getContentPane().setLayout(new BorderLayout());
	
	// Hide-show listener:
	
	this.addWindowListener(new WindowAdapter() {
		@Override public void windowClosing(WindowEvent e) {
			LiveGraph.application().setDisplayDataFileSettingsWindow(false);
		}
	});
	
	// Layout:	
	
	//JPanel panel = null;
	JButton button = null;
	Dimension dim = null;
	
	// Settings controls:
	
	JPanel settingsPanel = new JPanel(new GridBagLayout());	
	settingsPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
	getContentPane().add(settingsPanel, BorderLayout.NORTH);	
	settingsPanel.add(new Box.Filler((dim = new Dimension(1, 1)), dim, dim),
					  new GridBagConstraints(3, 0, 1, 1, 1, 0,
											 GridBagConstraints.WEST,
											 GridBagConstraints.BOTH,
											 new Insets(0, 0, 0, 0),
											 0, 0));
	
	// File name input:	
	
	settingsPanel.add(new JLabel("Data file:"), org.LiveGraph.gui.Tools.createGridBagConstraints(0, 0, 3, 1));
	
	openFileDialog = new JFileChooser();
	openFileDialog.addChoosableFileFilter(new FileFilter() {
        @Override public boolean accept(File f) {
        	if (null == f) return false;
        	if (f.isDirectory()) return true;
        	int p = f.getName().lastIndexOf(".");        
        	return p < 0 ? false : f.getName().substring(p).equalsIgnoreCase(".csv");
	    }
        @Override public String getDescription() { return "Comma separated values (*.csv)"; }
	});
	openFileDialog.addChoosableFileFilter(new FileFilter() {
		@Override public boolean accept(File f) {
        	if (null == f) return false;
        	if (f.isDirectory()) return true;
        	int p = f.getName().lastIndexOf(".");        
        	return p < 0 ? false : f.getName().substring(p).equalsIgnoreCase(".dat");
	    }
		@Override public String getDescription() { return "Generic data files (*.dat)"; }
	});
	openFileDialog.setCurrentDirectory(new File(System.getProperty("user.dir")));
	
	fileNameLabel = new JLabel("- no data file selected -");
	fileNameLabel.setFont(SwingTools.getPlainFont(fileNameLabel));
	settingsPanel.add(fileNameLabel, Tools.createGridBagConstraints(0, 1, 4, 1));
	settingsPanel.add((button = new JButton("Open...")), Tools.createGridBagConstraints(4, 1, 1, 1));
	button.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			
			if (JFileChooser.APPROVE_OPTION != openFileDialog.showOpenDialog(DATAFILE_WIN))
				return;
			if (!openFileDialog.getSelectedFile().exists())
				return;
			
			String filePath = openFileDialog.getSelectedFile().getAbsolutePath();
			try {
				LiveGraph.application().getDataFileSettings().setDataFile(filePath);
				LiveGraph.application().logSuccessLn("New source data file set: \"" + filePath + "\".");
				
			} catch (ErrorWhileSettingHasChangedProcessingException ex) {
								
				setFileNameLabel(null);
				String extraInfo = ".";
				if (null != ex.getCause())
					extraInfo = ": \n    (" + ex.getCause().getMessage() + ").";
					
				LiveGraph.application().logErrorLn("Error setting source data file to \"" + filePath + "\""
												 + extraInfo);												 
				
				try { LiveGraph.application().getDataFileSettings().setDataFile(""); }
				catch (Exception aex) {}
				LiveGraph.application().logInfoLn("Source data file re-set to: \"\".");
			}									
        }		
	});
	
	// Cache options:
	
	ButtonGroup bGroup = new ButtonGroup();
	bGroup.add(showAllDataButton = new JRadioButton("Show all data", true));
	bGroup.add(showTailDataButton = new JRadioButton("Show tail data", false));
	showAllDataButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			LiveGraph.application().getDataFileSettings().setShowOnlyTailData(false);
        }		
	});	
	showTailDataButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			LiveGraph.application().getDataFileSettings().setShowOnlyTailData(true);
        }		
	});
	settingsPanel.add(showAllDataButton, Tools.createGridBagConstraints(0, 3, 1, 1));
	settingsPanel.add(showTailDataButton, Tools.createGridBagConstraints(1, 3, 1, 1));
	
	dontCacheBox = new JCheckBox("Do not cache data", false);
	dontCacheBox.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			LiveGraph.application().getDataFileSettings().setDoNotCacheData(dontCacheBox.isSelected());			
        }		
	});
	settingsPanel.add(dontCacheBox, Tools.createGridBagConstraints(2, 3, 3, 1));
	
	// Update interval slider:
	
	settingsPanel.add(new JLabel("Update frequency:"), Tools.createGridBagConstraints(0, 4, 3, 1));
	updateIntervallSlider = new JSlider(0, updateIntervalLabels.length - 1, updateIntervalLabels.length - 1);
	updateIntervallSlider.setMinorTickSpacing(1);	
	updateIntervallSlider.setSnapToTicks(true);
	updateIntervallSlider.setPaintTicks(true);
	updateIntervallSlider.setPaintTrack(true);
	updateIntervallSlider.setPaintLabels(false);
	updateIntervallSlider.setMajorTickSpacing(1);
	updateIntervallSlider.addChangeListener(new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			int v = updateIntervallSlider.getValue();			
			LiveGraph.application().getDataFileSettings().setUpdateFrequency(updateIntervalValues[v]);
        }		
	});
	settingsPanel.add(updateIntervallSlider, Tools.createGridBagConstraints(0, 5, 5, 1));
	
	intervalLabel = new JLabel(updateIntervalLabels[updateIntervalLabels.length - 1]);
	intervalLabel.setFont(SwingTools.getPlainFont(intervalLabel));
	settingsPanel.add(intervalLabel, Tools.createGridBagConstraints(0, 6, 4, 1));
	
	// Update buttons & cache settings:

	nextUpdateLabel = new JLabel("Next update: manual. ");
	settingsPanel.add(nextUpdateLabel, Tools.createGridBagConstraints(0, 7, 4, 1));
	
	settingsPanel.add((button = new JButton("Update now")), Tools.createGridBagConstraints(4, 7, 1, 1));
	button.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			LiveGraph.application().initiateDataUpdate();
        }		
	});	
	
	
	// File info text field:
	
	this.fileInfoArea = new JTextArea();
	this.fileInfoArea.setEditable(false);
	JPanel fileInfoPanel = new JPanel(new BorderLayout(5, 5));
	fileInfoPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
	fileInfoPanel.add(new JLabel("File info:"), BorderLayout.NORTH);
	fileInfoPanel.add(new JScrollPane(this.fileInfoArea), BorderLayout.CENTER);
	getContentPane().add(fileInfoPanel, BorderLayout.CENTER);
	
}

/**
 * Updates an apropriate label as the invoker counts down the time until the next update.
 */
public void timerTick(UpdateInvoker source) {
	long remaining = source.getRemainingMillis();
	if (remaining < 0) {
		nextUpdateLabel.setText("Next update: on button click.");
		return;
	}
	
	long h = remaining / 3600000;
	long m = (remaining % 3600000) / 60000;
	double s = ((remaining % 3600000) % 60000) / 1000.;
	
	StringBuffer t = new StringBuffer("Next update: ");
	if (h > 0) {
		t.append(h);
		t.append(h == 1 ? " hour " : " hours ");
	}
	
	if (h > 0 || m > 0) {
		t.append(m);
		t.append(m == 1 ? " minute " : " minutes ");
	}
	
	t.append(s);
	t.append(" seconds.");
	
	nextUpdateLabel.setText(t.toString());
}

/**
 * Displays an apropriate label when the invoker has started an update from the data file.
 */
public void updateStarted(UpdateInvoker source) {
	nextUpdateLabel.setText("Update in progress.");	
}

/**
 * Displays an apropriate label when the invoker has finished an update from the data file. 
 */
public void updateFinished(UpdateInvoker source, String errorMsg) {
	
	if (null == errorMsg) {
		nextUpdateLabel.setText("Update finished successfully.");
		return;
	}
	
	String out = "Problem: " + errorMsg.trim();
	if (!out.endsWith("."))
		out = out + ".";
	
	nextUpdateLabel.setText(out);
}

/**
 * Sets the file name label in the window. If the label is too long, the baginning it stripped off.
 * @param fileName Data file name.
 */
private void setFileNameLabel(String fileName) {
	if (null == fileName || 0 == fileName.trim().length()) {
		fileNameLabel.setText("- no data file selected -");
		return;
	}
	fileName = fileName.trim();
	FontMetrics fm = fileNameLabel.getFontMetrics(fileNameLabel.getFont());
	if (fm.stringWidth(fileName) > fileNameLabel.getWidth() - 10) {
		while (fm.stringWidth("..." + fileName) > fileNameLabel.getWidth() - 10) {
			fileName = fileName.substring(1);	
		}
		fileName = "..." + fileName;
	}					
	fileNameLabel.setText(fileName);
}

/**
 * Updates the view when the settings were loaded from a file.
 * @param settings The settings.
 * @param info Event info.
 */
public void settingHasChanged(ObservableSettings settings, Object info) {
	
	if (null == settings)
		return;
	
	if (settings instanceof DataFileSettings) {
		settingHasChanged((DataFileSettings) settings, info);
		return;
	}
}

/**
 * Updates the view when the settings were loaded from a file.
 * @param settings The settings.
 * @param info Event info.
 */
public void settingHasChanged(DataFileSettings settings, Object info) {
	
	if (null == info || !(info instanceof String))
		return;
	
	String event = (String) info;
	
	if (event.equals("DataFile") || event.equals("load")) {
		setFileNameLabel(settings.getDataFile());
	}
	
	if (event.equals("ShowOnlyTailData") || event.equals("load")) {
		showAllDataButton.setSelected(!settings.getShowOnlyTailData());
		showTailDataButton.setSelected(settings.getShowOnlyTailData());
	}
	
	if (event.equals("DoNotCacheData") || event.equals("load")) {
		dontCacheBox.setSelected(settings.getDoNotCacheData());
	}
	
	if (event.equals("UpdateFrequency") || event.equals("load")) {		
		long f = settings.getUpdateFrequency();
		int p = (0 >= f ? updateIntervalValues.length - 1 : Arrays.binarySearch(updateIntervalValues, f));
		String lab;
		if (0 > p)
			lab = "every " + f + " milliseconds.";
		else
			lab = updateIntervalLabels[p];
		updateIntervallSlider.setValue(p);
		intervalLabel.setText(lab);		
	}	
}

/**
 * Displayes data file info.
 * @param text Info.
 */
private void setDataFileInfoText(String text) {
	fileInfoArea.setText(text + "\n ");
	JScrollBar sb = ((JScrollPane) fileInfoArea.getParent().getParent()).getVerticalScrollBar();
	if (null != sb)
		sb.setValue(sb.getMaximum());
}

/**
 * Updates data file info when the cache changes.
 */
public void cacheEventFired(DataCache cache, CacheEvent event) {
	
	switch(event) {
		case UpdateLabels:			
		case ChangeMode:
		case UpdateData:
			break;
		case UpdateDataFileInfo:
			setDataFileInfoText(cache.getDataFileInfo());
			break;
		default:
			throw new Error("This case is impossible!");
			
	}
}

}  
