package org.LiveGraph.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.WindowConstants;
import javax.swing.border.EtchedBorder;

import org.LiveGraph.LiveGraph;
import org.LiveGraph.dataCache.CacheObserver;
import org.LiveGraph.dataCache.DataCache;
import org.LiveGraph.plot.Plotter;
import org.LiveGraph.settings.DataSeriesSettings;
import org.LiveGraph.settings.GraphSettings;
import org.LiveGraph.settings.ObservableSettings;
import org.LiveGraph.settings.SettingsObserver;

import com.softnetConsult.utils.swing.SwingTools;


/**
 * The plot window of the application.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: PlotWindow.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class PlotWindow extends JFrame implements CacheObserver, SettingsObserver {

private JLabel statusLabel = null;
private Plotter plotter = null;
private JPanel canvas = null;
private List<SeriesHighlightListener> seriesHighlightListeners = null;
private boolean highlightDataPoints = true;
private Popup hlSerPopup = null; 
private List<String> seriesLabels = null;

/**
 * Created and setts up a plotter window.
 * 
 * @param plotter The plotter object for this window.
 */
public PlotWindow(Plotter plotter) {
	super();
	
	if (null == plotter)
		throw new NullPointerException("Cannot use a null plotter");
	
	this.seriesHighlightListeners = new ArrayList<SeriesHighlightListener>();
	this.plotter = plotter;
	this.highlightDataPoints = true;
	if (null != LiveGraph.application().getGraphSettings())
		this.highlightDataPoints = LiveGraph.application().getGraphSettings().getHighlightDataPoints();
	this.seriesLabels = new ArrayList<String>();
	this.initialize();
}

/**
 * This method initializes the window.
 */
private void initialize() {
	
	// Window settings:
	
	//final PlotWindow PLOT_WIN = this;
	this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
	Dimension frameDim = new Dimension(450, 500);
	this.setPreferredSize(frameDim);
	this.setBounds(480, 5, frameDim.width, frameDim.height);	
	this.setTitle("Data plot (LiveGraph)");
	
	// Hide-show listener:
	
	this.addWindowListener(new WindowAdapter() {
		@Override public void windowClosing(WindowEvent e) {
			LiveGraph.application().setDisplayPlotWindow(false);
		}
	});
	
	// Layout:
	
	JPanel panel = null;
	getContentPane().setLayout(new BorderLayout());
	
	
	// Ststus label:
	
	statusLabel = new JLabel();
	statusLabel.setFont(SwingTools.getPlainFont(statusLabel));
	
	panel = new JPanel(new BorderLayout());
	
	panel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(5, 3, 5, 5),
													   BorderFactory.createEtchedBorder(EtchedBorder.RAISED)));	
	panel.setPreferredSize(new Dimension(frameDim.width, 30));
	panel.add(statusLabel, BorderLayout.CENTER);
	
	getContentPane().add(panel , BorderLayout.SOUTH);
	
	// Plot canvas:	
	
	hlSerPopup = null; 
	canvas = new JPanel() {
		@Override public void paint(Graphics g) {
			super.paint(g);
			plotter.paint(g);
		}		
		@Override public void setBounds(int x, int y, int width, int height) {
			super.setBounds(x, y, width, height);
			plotter.setScreenSize(width, height);
		}
	};
	canvas.setBackground(Color.WHITE);
	canvas.addMouseListener(new MouseAdapter() {
		@Override public void mouseExited(MouseEvent e) {
			setStatusMessage("");
			if (highlightDataPoints) {
				List<Integer> l = Collections.emptyList();
				notifySeriesHighlightListeners(l);				
			}
		}
		@Override public void mouseReleased(MouseEvent e) {
			if (highlightDataPoints && MouseEvent.BUTTON1 == e.getButton() && null != hlSerPopup) {
				hlSerPopup.hide();
				hlSerPopup = null;
			}
		}
		@Override public void mousePressed(MouseEvent e) {
			if (highlightDataPoints && MouseEvent.BUTTON1 == e.getButton())
				showHlSeriesPopup(plotter.highlightAround(e.getPoint()), e.getXOnScreen(), e.getYOnScreen());
		}
	});
	canvas.addMouseMotionListener(new MouseMotionAdapter() {
		@Override public void mouseDragged(MouseEvent e) { this.mouseMoved(e); }
		@Override public void mouseMoved(MouseEvent e) {
			if (plotter.screenTooSmall()) {
				setStatusMessage("Enlarge this window.");
				return;
			}
			if (!plotter.getShowAtLeastOneSeries()) {
				setStatusMessage("No data selected for display.");
				return;
			} 
			Point ep = e.getPoint();
			Point2D.Double dp = plotter.screenToDataPoint(ep);
			setStatusMessage(String.format("(%.3f, %.3f)", dp.x, dp.y));
			if (highlightDataPoints) {
				List<Integer> hlSeries = plotter.highlightAround(ep);
				notifySeriesHighlightListeners(hlSeries);			
				canvas.repaint();
				if (MouseEvent.MOUSE_DRAGGED == e.getID() && null != hlSerPopup)
					showHlSeriesPopup(hlSeries, e.getXOnScreen(), e.getYOnScreen());
			}
		}
	});
	
	panel = new JPanel(new BorderLayout());
	panel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(5, 5, 3, 5),
			   										   BorderFactory.createEtchedBorder(EtchedBorder.RAISED)));
	panel.add(canvas, BorderLayout.CENTER);
	getContentPane().add(panel, BorderLayout.CENTER);
	
	
} // private void initialize()


/**
 * Shows the popup with the labels of highlighted data series.
 * 
 * @param hlSeries List of indices of highlighted data series.
 * @param mx Current mouse position on screen (x).
 * @param my Current mouse position on screen (y).
 */
private void showHlSeriesPopup(List<Integer> hlSeries, int mx, int my) {
		
	JPanel panel = new JPanel(new GridLayout(hlSeries.size(), 1, 2, 2));
	panel.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.RAISED));
	JLabel label = null;
	
	if (hlSeries.isEmpty()) {
		
		label = new JLabel("- no data series highlighted -");
		panel.add(label);
		
	} else {
		for (int s : hlSeries) {
			
			label = new JLabel(seriesLabels.get(s));
			label.setFont(SwingTools.getPlainFont(label));
			label.setForeground(LiveGraph.application().getDataSeriesSettings().getColour(s));
			panel.add(label);	
		}
	}
	
	
	if (null != hlSerPopup)
		hlSerPopup.hide();							
	hlSerPopup = PopupFactory.getSharedInstance().getPopup(canvas, panel, mx + 15, my + 15);
	hlSerPopup.show();
}

/**
 * Update the status bar message.
 * @param msg message.
 */
public void setStatusMessage(String msg) {
	if (null == msg)
		return;
	statusLabel.setText(msg);
}

/**
 * Repaints the plot canvas when the cache was updated.
 */
public void cacheEventFired(DataCache cache, CacheEvent event) {
	
	switch(event) {
		
		case UpdateData:
			canvas.repaint();
			break;
			
		case UpdateLabels:
			seriesLabels.clear();
			Iterator<String> labls = cache.iterateDataSeriesLabels();
			while (labls.hasNext())
				seriesLabels.add(labls.next());
			break;
		
		default:
			break;			
	}
} 

/**
 * Dispatches settings change events.
 * @see #settingHasChanged(DataSeriesSettings, String)
 * @see #settingHasChanged(GraphSettings, String)
 */
public void settingHasChanged(ObservableSettings settings, Object info) {
	
	if (null == info || null == settings)
		return;
	
	if ((settings instanceof DataSeriesSettings) && (info instanceof String)) {
		settingHasChanged((DataSeriesSettings) settings, (String) info);
		return;
	}
	
	if ((settings instanceof GraphSettings) && (info instanceof String)) {
		settingHasChanged((GraphSettings) settings, (String) info);
		return;
	}
}

/**
 * Repaints the plot canvas when the data file settings change.
 * 
 * @param settings Series settings (not currentluy used).
 * @param info Event info (not currently used).
 */
public void settingHasChanged(DataSeriesSettings settings, String info) {
	
	if (null == info || null == settings)
		return;
	
	canvas.repaint();
}

/**
 * Repaints the plot canvas when the graph settings change.
 * 
 * @param settings Series settings (not currentluy used).
 * @param info Event info (not currently used).
 */
public void settingHasChanged(GraphSettings settings, String info) {
	
	if (null == info || null == settings)
		return;
	
	if (info.equals("HighlightDataPoints") || info.equals("load")) {
		highlightDataPoints = settings.getHighlightDataPoints();
	}
	
	canvas.repaint();
}

/**
 * Adds a {@code SeriesHighlightListener}.
 * @param listener A listener.
 * @return Whether the listener was actually added because it did not exist.
 */
public boolean addSeriesHighlightListeners(SeriesHighlightListener listener) {
	return seriesHighlightListeners.add(listener);
}

/**
 * Removed a {@code SeriesHighlightListener}.
 * @param listener A listener.
 * @return Whether the listener was actually removed because it was there.
 */
public boolean removeSeriesHighlightListeners(SeriesHighlightListener listener) {
	return seriesHighlightListeners.remove(listener);
}

/**
 * @return Number of {@code SeriesHighlightListener}s.
 */
public int countSeriesHighlightListeners() {
	return seriesHighlightListeners.size();
}

/**
 * Notifies the listeners about the highlighted series.
 * @param hlSeries highlighted series indices.
 */
private void notifySeriesHighlightListeners(List<Integer> hlSeries) {
	for (SeriesHighlightListener listener : seriesHighlightListeners)
		listener.highlightSeries(hlSeries);
}


} // public class PlotWindow
